/* $Id: kAVLRemove2.c.h,v 1.2 2001/11/09 07:36:08 bird Exp $
 *
 * kAVLRemove2 - Remove specific node (by pointer) from an AVL tree.
 *
 * Copyright (c) 2001 knut st. osmundsen (knut.stange.osmundsen@mynd.no)
 *
 */

#ifndef _kAVLRemove2_h_
#define _kAVLRemove2_h_


/**
 * Removes a node from the AVL-tree given by it's pointer.
 * @returns   Pointer to the node.
 * @param     ppTree  Pointer to the AVL-tree root node pointer.
 * @param     Key     Key value of the node which is to be removed.
 * @sketch    Find the node which is to be removed:
 *            LOOP until not found
 *            BEGIN
 *                Add node pointer pointer to the AVL-stack.
 *                IF the keys matches THEN break!
 *                IF remove key < node key THEN
 *                    left
 *                ELSE
 *                    right
 *            END
 *            IF found THEN
 *            BEGIN
 *                IF duplicate keys allowed AND this isn't pNode Then
 *                BEGIN
 *                    Search list (pList) for the node.
 *                    IF found THEN Link it out and return it.
 *                    return NULL (if not found).
 *                END
 *                IF duplicate keys allowed AND this is pNode AND there is nodes in the list Then
 *                    Replace pNode with the first node in the list.
 *                END
 *                IF left node not empty THEN
 *                BEGIN
 *                    Find the right most node in the left tree while adding the pointer to the pointer to it's parent to the stack:
 *                    Start at left node.
 *                    LOOP until right node is empty
 *                    BEGIN
 *                        Add to stack.
 *                        go right.
 *                    END
 *                    Link out the found node.
 *                    Replace the node which is to be removed with the found node.
 *                    Correct the stack entry for the pointer to the left tree.
 *                END
 *                ELSE
 *                BEGIN
 *                    Move up right node.
 *                    Remove last stack entry.
 *                END
 *                Balance tree using stack.
 *            END
 *            return pointer to the removed node (if found).
 * @status    completely implemented.
 * @author    knut st. osmundsen
 */
PAVLNODECORE KAVL_FN(Remove2)(PPAVLNODECORE ppTree, PAVLNODECORE pNode)
{
    KLOGENTRY2("PAVLNODECORE","PPAVLNODECORE ppTree, PAVLNODECORE pNode", ppTree, pNode);
    AVLKEY                  Key = pNode->Key;
    AVLSTACK                AVLStack;
    PPAVLNODECORE           ppDeleteNode = ppTree;
    register PAVLNODECORE   pDeleteNode;

    AVLStack.cEntries = 0;

    while ((pDeleteNode = *ppDeleteNode) != NULL)
    {
        kASSERT(AVLStack.cEntries < KAVL_MAX_STACK);
        AVLStack.aEntries[AVLStack.cEntries++] = ppDeleteNode;
        #ifndef AVL_CMP
        if (AVL_E(pDeleteNode->Key, Key))
            break;

        if (AVL_G(pDeleteNode->Key, Key))
            ppDeleteNode = &pDeleteNode->pLeft;
        else
            ppDeleteNode = &pDeleteNode->pRight;
        #else
        {
        int register iDiff;
        if ((iDiff = AVL_CMP(pDeleteNode->Key, Key)) == 0)
            break;

        if (iDiff > 0)
            ppDeleteNode = &pDeleteNode->pLeft;
        else
            ppDeleteNode = &pDeleteNode->pRight;
        }
        #endif
    }

    if (pDeleteNode != NULL)
    {
        #ifdef KAVL_EQUAL_ALLOWED
        if (pDeleteNode != pNode)
        {
            while (pDeleteNode->pList != NULL)
            {
                if (pDeleteNode->pList == pNode)
                {
                    pDeleteNode->pList = pNode->pList;
                    pNode->pList = NULL;                                        //debug only?
                    KLOGEXIT(pNode);
                    return pNode;
                }
                pDeleteNode = pDeleteNode->pList;
            }
            KLOGEXIT(NULL);
            return NULL;
        }

        if (pDeleteNode->pList != NULL)
        {
            pDeleteNode = pDeleteNode->pList;
            pDeleteNode->pLeft = pNode->pLeft;
            pDeleteNode->pRight = pNode->pRight;
            pDeleteNode->uchHeight = pNode->uchHeight;
            kASSERT(*ppDeleteNode == pNode);
            *ppDeleteNode = pDeleteNode;
            pNode->pList = pNode->pLeft = pNode->pRight = NULL;                 //debug only?
            KLOGEXIT(pNode);
            return pNode;
        }
        #endif

        if (pDeleteNode->pLeft != NULL)
        {
            unsigned                iStackEntry = AVLStack.cEntries;
            PPAVLNODECORE           ppLeftLeast = &pDeleteNode->pLeft;
            register PAVLNODECORE   pLeftLeast;

            while ((pLeftLeast = *ppLeftLeast)->pRight != NULL) /* Left most node. */
            {
                kASSERT(AVLStack.cEntries < KAVL_MAX_STACK);
                AVLStack.aEntries[AVLStack.cEntries++] = ppLeftLeast;
                ppLeftLeast = &pLeftLeast->pRight;
                pLeftLeast  = pLeftLeast->pRight;
            }

            /* link out pLeftLeast */
            *ppLeftLeast = pLeftLeast->pLeft;

            /* link in place of the delete node. */
            pLeftLeast->pLeft = pDeleteNode->pLeft;
            pLeftLeast->pRight = pDeleteNode->pRight;
            pLeftLeast->uchHeight = pDeleteNode->uchHeight;
            *ppDeleteNode = pLeftLeast;
            AVLStack.aEntries[iStackEntry] = &pLeftLeast->pLeft;
        }
        else
        {
            *ppDeleteNode = pDeleteNode->pRight;
            AVLStack.cEntries--;
        }

        KAVL_FN(Rebalance)(SSToDS(&AVLStack));
    }

    #ifndef KAVL_EQUAL_ALLOWED                                  //debug only?
    pNode->pLeft = pNode->pRight = NULL;                        //debug only?
    #else                                                       //debug only?
    pNode->pList = pNode->pLeft = pNode->pRight = NULL;         //debug only?
    #endif                                                      //debug only?
    KLOGEXIT(pDeleteNode);
    return pDeleteNode;
}

#endif
