/* $Id: kAVLRemoveBestFit.c.h,v 1.2 2001/11/09 07:36:08 bird Exp $
 *
 * kAVLRemoveBestFit - Remove Best Fit routine for AVL trees.
 *                     Intended specially on heaps. The tree should allow duplicate keys.
 *
 * Copyright (c) 1999-2001 knut st. osmundsen (knut.stange.osmundsen@mynd.no)
 *
 */

#ifndef _kAVLRemoveBestFit_h_
#define _kAVLRemoveBestFit_h_


/**
 * Finds the best fitting node in the tree for the given Key value.
 * And removes it.
 * @returns   Pointer to the best fitting node found.
 * @param     ppTree  Pointer to Pointer to the tree root node.
 * @param     Key     The Key of which is to be found a best fitting match for..
 * @param     fAbove  TRUE:  Returned node is have the closest key to Key from above.
 *                    FALSE: Returned node is have the closest key to Key from below.
 * @status    completely implemented.
 * @sketch    The best fitting node is always located in the searchpath above you.
 *            >= (above): The node where you last turned left.
 *            <= (below): the node where you last turned right.
 * @author    knut st. osmundsen
 * @remark    This implementation should be speeded up slightly!
 */
PAVLNODECORE    KAVL_FN(RemoveBestFit)(PPAVLNODECORE ppTree, AVLKEY Key, KBOOL fAbove)
{
    KLOGENTRY3("PAVLNODECORE","PPAVLNODECORE ppTree, AVLKEY Key, KBOOL fAbove", ppTree, Key, fAbove);
    #ifdef AVL_CMP
    register int            iDiff;
    #endif
    register PAVLNODECORE   pNode = *ppTree;
    PAVLNODECORE            pNodeLast = NULL;

    if (fAbove)
    {   /* pNode->Key >= Key */
        #ifndef AVL_CMP
        while (pNode != NULL && AVL_NE(pNode->Key, Key))
        #else
        while (pNode != NULL && (iDiff = AVL_CMP(pNode->Key, Key)) != 0)
        #endif
        {
            #ifndef AVL_CMP
            if (AVL_G(pNode->Key, Key))
            #else
            if (iDiff > 0)
            #endif
            {
                pNodeLast = pNode;
                pNode = pNode->pLeft;
            }
            else
                pNode = pNode->pRight;
        }
    }
    else
    {   /* pNode->Key <= Key */
        #ifndef AVL_CMP
        while (pNode != NULL && AVL_NE(pNode->Key, Key))
        #else
        while (pNode != NULL && (iDiff = AVL_CMP(pNode->Key, Key)) != 0)
        #endif
        {
            #ifndef AVL_CMP
            if (AVL_L(pNode->Key, Key))
            #else
            if (iDiff < 0)
            #endif
            {
                pNodeLast = pNode;
                pNode = pNode->pRight;
            }
            else
                pNode = pNode->pLeft;
        }
    }

    /*
     * If we found anything we'll have to remove the node and return it.
     * If duplicate keys are allowed we'll check for multiple nodes first
     * and return one of them before doing an expensive remove operation
     * of the node we found.
     */
    if (pNode == NULL)
        pNode = pNodeLast;
    if (pNode == NULL)
    {
        KLOGEXIT(NULL);
        return NULL;
    }
    #ifdef KAVL_EQUAL_ALLOWED
    if (pNode->pList)
    {
        pNodeLast = pNode->pList;
        pNode->pList = pNodeLast->pList;
        KLOGEXIT(pNodeLast);
        return pNodeLast;
    }
    #endif
    pNode = KAVL_FN(Remove)(ppTree, pNode->Key);
    KLOGEXIT(pNode);
    return pNode;
}


#endif
